import { resolve, relative } from 'path';
import * as process from 'process';
import * as util from 'util';
import { createCache } from '@file-cache/core';
import { createNpmPackageKey } from '@file-cache/npm';
import AwaitLock from 'await-lock';
import chalk from 'chalk';
import * as chokidar from 'chokidar';
import _glob from 'glob';
import { DEFAULT_ARBITRARY_EXTENSIONS } from './config.js';
import { isGeneratedFilesExist, emitGeneratedFiles } from './emitter/index.js';
import { Locator } from './locator/index.js';
import { Logger } from './logger.js';
import { createDefaultResolver } from './resolver/index.js';
import { createDefaultTransformer } from './transformer/index.js';
import { getInstalledPeerDependencies, isMatchByGlob } from './util.js';
const glob = util.promisify(_glob);
export async function run(options) {
    const lock = new AwaitLock.default();
    const logger = new Logger(options.logLevel ?? 'info');
    const cwd = options.cwd ?? process.cwd();
    const resolver = options.resolver ??
        createDefaultResolver({
            cwd,
            sassLoadPaths: options.sassLoadPaths,
            lessIncludePaths: options.lessIncludePaths,
            webpackResolveAlias: options.webpackResolveAlias,
        });
    const transformer = options.transformer ?? createDefaultTransformer({ cwd, postcssConfig: options.postcssConfig });
    const installedPeerDependencies = await getInstalledPeerDependencies();
    const cache = await createCache({
        mode: options.cacheStrategy ?? 'content',
        keys: [
            () => createNpmPackageKey(['happy-css-modules', ...installedPeerDependencies]),
            () => {
                return JSON.stringify(options);
            },
        ],
        noCache: !(options.cache ?? true),
    });
    const locator = new Locator({ transformer, resolver });
    const isExternalFile = (filePath) => {
        return !isMatchByGlob(filePath, options.pattern, { cwd });
    };
    async function processFile(filePath) {
        async function isChangedFile(filePath) {
            const result = await cache.getAndUpdateCache(filePath);
            if (result.error)
                throw result.error;
            return result.changed;
        }
        // Locator#load cannot be called concurrently. Therefore, it takes a lock and waits.
        await lock.acquireAsync();
        try {
            const _isGeneratedFilesExist = await isGeneratedFilesExist(filePath, options.declarationMap, options.arbitraryExtensions ?? DEFAULT_ARBITRARY_EXTENSIONS);
            const _isChangedFile = await isChangedFile(filePath);
            // Generate .d.ts and .d.ts.map only when the file has been updated.
            // However, if .d.ts or .d.ts.map has not yet been generated, always generate.
            if (_isGeneratedFilesExist && !_isChangedFile) {
                logger.debug(chalk.gray(`${relative(cwd, filePath)} (skipped)`));
                return;
            }
            const result = await locator.load(filePath);
            await emitGeneratedFiles({
                filePath,
                tokens: result.tokens,
                emitDeclarationMap: options.declarationMap,
                dtsFormatOptions: {
                    localsConvention: options.localsConvention,
                    arbitraryExtensions: options.arbitraryExtensions,
                },
                isExternalFile,
            });
            logger.info(chalk.green(`${relative(cwd, filePath)} (generated)`));
            await cache.reconcile(); // Update cache for the file
        }
        finally {
            lock.release();
        }
    }
    async function processAllFiles() {
        const filePaths = (await glob(options.pattern, { dot: true, cwd }))
            // convert relative path to absolute path
            .map((file) => resolve(cwd, file));
        const errors = [];
        for (const filePath of filePaths) {
            await processFile(filePath).catch((e) => errors.push(e));
        }
        if (errors.length > 0) {
            throw new AggregateError(errors, 'Failed to process files');
        }
    }
    if (!options.watch) {
        logger.info('Generate .d.ts for ' + options.pattern + '...');
        await processAllFiles();
        // Write cache state to file for persistence
    }
    else {
        // First, watch files.
        logger.info('Watch ' + options.pattern + '...');
        const watcher = chokidar.watch([options.pattern.replace(/\\/g, '/')], { cwd });
        watcher.on('all', (eventName, relativeFilePath) => {
            const filePath = resolve(cwd, relativeFilePath);
            // There is a bug in chokidar that matches symlinks that do not match the pattern.
            // ref: https://github.com/paulmillr/chokidar/issues/967
            if (isExternalFile(filePath))
                return;
            if (eventName !== 'add' && eventName !== 'change')
                return;
            processFile(filePath).catch((e) => {
                logger.error(e);
                // TODO: Emit a error by `Watcher#onerror`
            });
        });
        return { close: async () => watcher.close() };
    }
}
//# sourceMappingURL=runner.js.map